import os
import json
import joblib
import pickle


def load_model(model_path) -> object:
    """Load a model from the specified path.

    Args:
        model_path: Path to the model file.
    """
    if not os.path.exists(model_path):
        raise FileNotFoundError(f"No file found at {model_path}")
    ext = os.path.splitext(model_path)[1].lower()

    if ext == ".pickle":
        with open(model_path, "rb") as f:
            return pickle.load(f)

    elif ext == ".json":
        with open(model_path, "r") as f:
            return json.load(f)

    elif ext == ".joblib":
        return joblib.load(model_path)

    else:
        raise ValueError(f"Unsupported model_type: {ext}")



from typing import List, Optional, Union
import pandas as pd
from .model import Forecaster, Regressor
from .graph import EffectDirection, PerturbationDirection
from .metrics import MetricsCalculator
import graphviz as gr


class Verifier:
    """Class to verify the causal effect given trained models and datasets."""

    def __init__(self,
                 model: Union[Forecaster, Regressor],
                 original_dataset: pd.DataFrame,
                 perturbed_dataset: pd.DataFrame,
                 covariates: List[str],
                 perturbation_direction: PerturbationDirection,
                 effect_direction: EffectDirection,
                 scaled: bool = True):
        """
        Args:
            model: Model class
            original_dataset (pd.DataFrame): Original dataset
            perturbed_dataset (pd.DataFrame): Dataset having covariate(s) perturbed
            covariates (List[str]): Name(s) of specified covariate(s)
            perturbation_direction (PerturbationDirection): Direction of perturbation process (increasing or decreasing)
            effect_direction (EffectType): Relationship between covariate(s) and target forecast (positive or negative)

        """
        self.model = model
        self.original_dataset = original_dataset
        self.perturbed_dataset = perturbed_dataset
        self.covariates = covariates
        self.perturbation_direction = perturbation_direction
        self.effect_direction = effect_direction
        self.scaled = scaled

        # Check if original and perturbed datasets have the same shape
        assert self.original_dataset.shape == self.perturbed_dataset.shape, "Both datasets should have same shape"

    def assumption_plot(self, render: bool = False):
        # visualize covariate causal graph using pygraphviz
        g = gr.Digraph()
        g.node("Target", shape="box")
        for covar in self.covariates:
            g.node(covar)
            g.edge(covar, "Target", label=self.effect_direction.name)
        if render:
            g.render("covariate_causal_graph", format="png", cleanup=True)
        return g

    def procedure_plot(self, render: bool = False):
        g = gr.Digraph()
        g.node("Original Dataset", shape="box")
        g.node("Perturbed Dataset", shape="box")
        g.node("Model", shape="box")
        g.node("Predictions", shape="box")
        g.node("Predictions Perturbed", shape="box")
        g.node("Violation score", shape="box")
        g.edge("Original Dataset", "Perturbed Dataset",
               label=str(self.perturbation_direction))
        g.edge("Original Dataset", "Model")
        g.edge("Perturbed Dataset", "Model")
        g.edge("Model", "Predictions")
        g.edge("Model", "Predictions Perturbed")
        g.edge("Predictions", "Violation score")
        g.edge("Predictions Perturbed", "Violation score")

        if render:
            g.render("procedure_graph", format="png", cleanup=True)
        return g

    def _check_covariate_direction_violation(self):

        # check original_dataset == perturbed_dataset except for covariate column
        if not self.original_dataset.drop(columns=self.covariates).\
                equals(self.perturbed_dataset.drop(columns=self.covariates)):
            return True
        # and check if covariate is perturbed in the specified direction
        if self.perturbation_direction == PerturbationDirection.increasing:
            return (self.original_dataset[self.covariates] > self.perturbed_dataset[self.covariates]).any().sum()
        elif self.perturbation_direction == PerturbationDirection.decreasing:
            return (self.original_dataset[self.covariates] < self.perturbed_dataset[self.covariates]).any().sum()
        return True

    def __call__(self) -> dict:
        if not self.scaled and self._check_covariate_direction_violation():
            raise ValueError("Covariate direction check failed.")
        predictions = self.model.forecast(self.original_dataset) if isinstance(
            self.model, Forecaster) else self.model.predict(self.original_dataset)
        predictions_perturbed = self.model.forecast(self.perturbed_dataset) if isinstance(
            self.model, Forecaster) else self.model.predict(self.perturbed_dataset)

        alignment_score = MetricsCalculator.alignment_score(predictions,
                                                            predictions_perturbed,
                                                            perturbation_direction=self.perturbation_direction,
                                                            effect_direction=self.effect_direction)

        result = {
            "assumption": f"{self.covariates} → target",
            "relationship": f"{self.effect_direction}",
            "perturbation_direction": self.perturbation_direction,
            "alignment_score": alignment_score,
            "target_length": predictions.shape[-1]
        }
        return result



from .verifier import Verifier
from .metrics import MetricsCalculator
from .model import Regressor, Forecaster

__all__ = ["Verifier",
           "MetricsCalculator",
           "Regressor",
           "Forecaster"
           ]
__version__ = "0.1.2"



from enum import Enum
from typing import Optional, Union, Tuple
import warnings
from abc import ABCMeta
import pandas as pd
import numpy as np
from .graph import PerturbationDirection


class PerturbationType(Enum):
    """Enum-like class for perturbation functions."""
    ADDITIVE = "additive"
    MULTIPLICATIVE = "multiplicative"
    BINARY = "binary"
    SAMPLE = "sample"


class PerturbationDataset():

    def __init__(self,
                 original_dataset: pd.DataFrame,
                 covariates: list[str],
                 perturbation_direction: PerturbationDirection):
        self._perturbation_dataset = original_dataset.copy()
        self._covariates = covariates
        self._direction = perturbation_direction.value

    def additive_perturb(self, const: float,
                         bound: Optional[Tuple[float, float]] = None
                         ):

        for covar in self._covariates:
            self._perturbation_dataset[covar] = self._perturbation_dataset[covar] + \
                const * self._direction
        if bound is not None:
            min_value = bound[0]
            max_value = bound[1]
            if min_value >= max_value:
                raise ValueError(
                    "Not valid bound, min value has to be smaller than max value")

            # add const within bound, else get closet value of min or max
            for covar in self._covariates:
                self._perturbation_dataset[covar] = self._perturbation_dataset[covar].clip(
                    lower=min_value, upper=max_value)

        return self._perturbation_dataset

    def multipicative_perturb(self, const: float,
                              bound: Optional[Tuple[float, float]] = None):

        for covar in self._covariates:
            self._perturbation_dataset[covar] = self._perturbation_dataset[covar] * \
                (1 + const * self._direction)

        if bound is not None:
            min_value = bound[0]
            max_value = bound[1]
            if min_value >= max_value:
                raise ValueError(
                    "Not valid bound, min value has to be smaller than max value")

            # add const within bound, else get closet value of min or max
            for covar in self._covariates:
                self._perturbation_dataset[covar] = self._perturbation_dataset[covar].clip(
                    lower=min_value, upper=max_value)

        return self._perturbation_dataset

    def binary_perturb(self):
        self._perturbation_dataset_binary_list = []
        if not self._direction == 0:
            raise ValueError(
                'User has to specified Perubation Direction to be binary')
        for covar in self._covariates:
            unique_values = self._perturbation_dataset[covar].unique().tolist()
            if len(unique_values) > 2:
                warnings.warn(
                    f"Categorical covariates provided, number of unique values: {len(unique_values)}")
            for value in unique_values:
                self._perturbation_dataset_binary = self._perturbation_dataset.copy()
                self._perturbation_dataset_binary[covar] = value
                self._perturbation_dataset_binary_list.append(
                    self._perturbation_dataset_binary)
        return self._perturbation_dataset_binary_list

    def add_gaussian_noise(self, mu: float, sigma: float, bound: Tuple[float, float] = None):
        for covar in self._covariates:
            self._perturbation_dataset[covar] = self._perturbation_dataset[covar].apply(
                lambda x: x + abs(np.random.normal(mu, sigma)) * self._direction)
        if bound is not None:
            min_value = bound[0]
            max_value = bound[1]
            if min_value >= max_value:
                raise ValueError(
                    "Not valid bound, min value has to be smaller than max value")
            for covar in self._covariates:
                self._perturbation_dataset[covar] = self._perturbation_dataset[covar].clip(
                    lower=min_value, upper=max_value)
        return self._perturbation_dataset



"""
Function to read electricity market data either locally or from an online database
"""

# Author: Jesus Lago

# License: AGPL-3.0 License


import pandas as pd
import os
import holidays

def read_data(path, dataset='PJM', years_test=2, begin_test_date=None, end_test_date=None, Select_Inputs='All'):
    """Function to read and import data from day-ahead electricity markets. 
    
    It receives a ``dataset`` name, and the ``path`` of the folder where datasets are saved. 
    It reads the file ``dataset.csv`` in the ``path`` directory and provides a split between training and
    testing dataset based on the test dates provided.

    It also names the columns of the training and testing dataset to match the requirements of the
    prediction models of the library. Namely, assuming that there are `N` exogenous inputs,
    the columns of the resulting training and testing dataframes are named
    ``['Price', 'Exogenous 1', 'Exogenous 2', ...., 'Exogenous N']``.

    If `dataset` is either ``"PJM"``, ``"NP"``, ``"BE"``, ``"FR"``, or ``"DE"``,
    the function checks whether ``dataset.csv`` exists in ``path``. If it doesn't exist,
    it downloads the data from an online database and saves it under the ``path`` directory. ``"PJM"``
    refes to the Pennsylvania-New Jersey-Maryland market, ``"NP"`` to the Nord Pool market,
    and ``"BE"``, ``"FR"``, and ``"DE"`` respectively to the EPEX-Belgium, EPEX-France, and EPEX-Germany 
    day-ahead markets.

    Note that the data available online for these five markets is limited to certain periods (see the 
    `database <https://zenodo.org/records/4624805>`_ for further details).  
    
    Parameters
    ----------
    path : str, optional
        Path where the datasets are stored or, if they do not exist yet, the path where the datasets 
        are to be stored
    nlayers : int, optional
        Number of hidden layers in the neural network
    dataset : str, optional
        Name of the dataset/market under study. If it is one one of the standard markets, 
        i.e. ``"PJM"``, ``"NP"``, ``"BE"``, ``"FR"``, or ``"DE"``, the dataset is automatically downloaded. If the name
        is different, a dataset with a csv format should be place in the ``path``.
    years_test : int, optional
        Number of years (a year is 364 days) in the test dataset. It is only used if 
        the arguments begin_test_date and end_test_date are not provided.
    begin_test_date : datetime/str, optional
        Optional parameter to select the test dataset. Used in combination with the argument
        ``end_test_date``. If either of them is not provided, the test dataset is built using the 
        ``years_test`` argument. ``begin_test_date`` should either be a string with the following 
        format ``"%d/%m/%Y %H:%M"``, or a datetime object.
    end_test_date : datetime/str, optional
        Optional parameter to select the test dataset. Used in combination with the argument
        ``begin_test_date``. If either of them is not provided, the test dataset is built using the 
        ``years_test`` argument. ``end_test_date`` should either be a string with the following 
        format ``"%d/%m/%Y %H:%M"``, or a datetime object.       
    Returns
    -------
    pandas.DataFrame, pandas.DataFrame
        Training dataset, testing dataset

    Example
    --------
    >>> from epftoolbox.data import read_data
    >>> df_train, df_test = read_data(path='.', dataset='PJM', begin_test_date='01-01-2016', 
    ...                               end_test_date='01-02-2016')
    Test datasets: 2016-01-01 00:00:00 - 2016-02-01 23:00:00
    >>> df_train.tail()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2015-12-31 19:00:00  29.513832     100700.0      13015.0
    2015-12-31 20:00:00  28.440134      99832.0      12858.0
    2015-12-31 21:00:00  26.701700      97033.0      12626.0
    2015-12-31 22:00:00  23.262253      92022.0      12176.0
    2015-12-31 23:00:00  22.262431      86295.0      11434.0
    >>> df_test.head()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2016-01-01 00:00:00  20.341321      76840.0      10406.0
    2016-01-01 01:00:00  19.462741      74819.0      10075.0
    2016-01-01 02:00:00  17.172706      73182.0       9795.0
    2016-01-01 03:00:00  16.963876      72300.0       9632.0
    2016-01-01 04:00:00  17.403722      72535.0       9566.0
    >>> df_test.tail()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2016-02-01 19:00:00  28.056729      99400.0      12680.0
    2016-02-01 20:00:00  26.916456      97553.0      12495.0
    2016-02-01 21:00:00  24.041505      93983.0      12267.0
    2016-02-01 22:00:00  22.044896      88535.0      11747.0
    2016-02-01 23:00:00  20.593339      82900.0      10974.0

    """

    # Checking if provided directory exist and if not create it
    if not os.path.exists(path):
        os.makedirs(path)

    # If dataset is one of the existing open-access ones,
    # they are imported if they exist locally or download from 
    # the repository if they do not
 
    try:
        file_path = os.path.join(path, dataset + '.csv')
        data_aux = pd.read_csv(file_path, index_col=0)
    except IOError as e:
        raise IOError("%s: %s" % (path, e.strerror))

    data_aux.index = pd.to_datetime(data_aux.index, dayfirst=True)
    selected_columns = ['Price'] + Select_Inputs
    data = data_aux[selected_columns].copy()

    columns = ['Price']
    n_exogeneous_inputs = len(data.columns) - 1

    for n_ex in range(1, n_exogeneous_inputs + 1):
        columns.append('Exogenous ' + str(n_ex))
        
    data.columns = columns

    # The training and test datasets can be defined by providing a number of years for testing
    # or by providing the init and end date of the test period
    if begin_test_date is None and end_test_date is None:
        number_datapoints = len(data.index)
        number_training_datapoints = number_datapoints - 24 * 364 * years_test

        # We consider that a year is 52 weeks (364 days) instead of the traditional 365
        df_train = data.loc[:data.index[0] + pd.Timedelta(hours=number_training_datapoints - 1), :]
        df_test = data.loc[data.index[0] + pd.Timedelta(hours=number_training_datapoints):, :]
    
    else:
        try:
            begin_test_date = pd.to_datetime(begin_test_date, dayfirst=True)
            end_test_date = pd.to_datetime(end_test_date, dayfirst=True)
        except ValueError:
            print("Provided values for dates are not valid")

        if begin_test_date.hour != 0:
            raise Exception("Starting date for test dataset should be midnight") 
        if end_test_date.hour != 23:
            if end_test_date.hour == 0:
                end_test_date = end_test_date + pd.Timedelta(hours=23)
            else:
                raise Exception("End date for test dataset should be at 0h or 23h") 

        print('Test datasets: {} - {}'.format(begin_test_date, end_test_date))
        df_train = data.loc[:begin_test_date - pd.Timedelta(hours=1), :]
        df_test = data.loc[begin_test_date:end_test_date, :]

    return df_train, df_test



from dataclasses import dataclass, field
from typing import Literal, Optional, Any, Protocol, Union, runtime_checkable
from enum import Enum
import pandas as pd
import numpy as np


class ForecasterType(str, Enum):
    point = "point"
    quantile = "quantile"
    density = "density"
    ensemble = "ensemble"


@dataclass
class AuthorInfo:
    """Author information.

    Attributes:
        name: Name of the author.
        email: Email of the author.
    """

    name: str
    email: Optional[str] = None


@dataclass
class ModelInfo:
    """Model information.

    Attributes:
        name: Name of the model.
        authors: List of authors.
        type: Type of the model.
        params: Parameters of the model.
    """

    name: str
    authors: list[AuthorInfo]
    # type: ForecasterType
    # sk_forecaster_type: Optional[SKForecastType] = None
    params: dict[str, Any] = field(default_factory=dict)


@runtime_checkable
class Regressor(Protocol):
    def info(self) -> ModelInfo: ...

    def predict(self, dataset: pd.DataFrame) -> pd.Series:
        """
        Predict function.
        Parameters
        ----------
        dataset : pandas dataframe
            test/evaluation dataset for model to make predictions.

        Returns
        -------
        predictions : pandas Series
            Predicted values.

        """
        pass


@runtime_checkable
class Forecaster(Protocol):
    def info(self) -> ModelInfo: ...

    def forecast(self,
                 steps: int,
                 last_window: Union[pd.Series, pd.DataFrame] = None,
                 exog: Union[pd.Series, pd.DataFrame] = None) -> Union[pd.Series, np.ndarray]:
        """
        Predict n steps ahead.
        Rererences: skforecast.base._forecaster_base.py
        Parameters
        ----------
        steps : int
            Number of steps to predict. 
        last_window : pandas Series, pandas DataFrame, default None
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.

        Returns
        -------
        predictions : pandas Series
            Predicted values.

        """
        pass



from enum import Enum
from typing import Callable, Protocol, Union
import pandas as pd


class EffectDirection(Enum):
    positive = 1
    negative = -1
    # bounded = "bounded"


class PerturbationDirection(Enum):
    increasing = 1  # X increase
    decreasing = -1  # X decrease
    binary = 0



import pandas as pd
import numpy as np
from spicy import stats
from typing import Dict, Union
from .graph import EffectDirection, PerturbationDirection


class MetricsCalculator:
    """Class to calculate various metrics for forecast verification."""

    @staticmethod
    def calculate_mae(actual: pd.Series, forecast: pd.Series) -> float:
        """Calculates Mean Absolute Error (MAE)."""
        return np.mean(np.abs(actual - forecast))

    @staticmethod
    def calculate_rmse(actual: pd.Series, forecast: pd.Series) -> float:
        """Calculates Root Mean Squared Error (RMSE)."""
        return np.sqrt(np.mean((actual - forecast) ** 2))

    @staticmethod
    def calculate_mape(actual: pd.Series, forecast: pd.Series) -> float:
        """Calculates Mean Absolute Percentage Error (MAPE)."""
        return np.mean(np.abs((actual - forecast) / actual)) * 100

    @staticmethod
    def t_statistics(series_old: pd.Series, series_new: pd.Series) -> Dict[str, float]:
        """Calculates t-statistics for two series."""
        if len(series_old) != len(series_new):
            raise ValueError(
                "Both series must have the same length for t-statistics calculation.")
        if series_old.empty or series_new.empty:
            raise ValueError(
                "Both series must not be empty for t-statistics calculation.")
        if not isinstance(series_old, pd.Series) or not isinstance(series_new, pd.Series):
            raise TypeError(
                "Both inputs must be pandas Series for t-statistics calculation.")
        if series_old.isnull().any() or series_new.isnull().any():
            raise ValueError(
                "Both series must not contain NaN values for t-statistics calculation.")
        tstats = stats.ttest_ind(series_old, series_new)
        """Returns a t statistics with 2 series"""
        return {
            "t_statistic": tstats.statistic,
            "p_value": tstats.pvalue
        }

    @staticmethod
    def alignment_score(pred_original: Union[np.ndarray, pd.Series],
                        pred_perturbed: Union[np.ndarray, pd.Series],
                        effect_direction: EffectDirection,
                        perturbation_direction: PerturbationDirection) -> float:

        assert pred_original.shape == pred_perturbed.shape, "Both series have to be the same length"
        expected = perturbation_direction.value * effect_direction.value
        alignment = (expected) * pred_perturbed >= (expected) * pred_original
        alignment_score = alignment.ravel().sum() / pred_original.shape[-1]

        return alignment_score



"""
Classes and functions to implement data wrangling operations. At the moment, this is limited to
data scaling.
"""

# Author: Jesus Lago

# License: AGPL-3.0 License

import numpy as np
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from statsmodels.robust import mad

class MedianScaler(object):

    def __init__(self):
        self.fitted = False

    def fit(self, data):

        if len(data.shape)!=2:
            raise IndexError('Error: Provide 2-D array. First dimension is datapoints and' + 
                  ' second features')
            return -1

        self.median = np.median(data, axis=0)
        self.mad = mad(data, axis=0)
        self.fitted = True
        
    def fit_transform(self, data):

        self.fit(data)
        return self.transform(data)
    
    def transform(self, data):

        if not self.fitted:
            print('Error: The scaler has not been yet fitted. Called fit or fit_transform')
            return -1

        if len(data.shape)!=2:
            raise IndexError('Error: Provide 2-D array. First dimension is datapoints and' + 
                  ' second features')

        transformed_data = np.zeros(shape=data.shape)

        for i in range(data.shape[1]):

            transformed_data[:, i] = (data[:, i] - self.median[i]) / self.mad[i]

        return transformed_data

    def inverse_transform(self, data):

        if not self.fitted:
            print('Error: The scaler has not been yet fitted. Called fit or fit_transform')
            return -1

        if len(data.shape)!=2:
            raise IndexError('Error: Provide 2-D array. First dimension is datapoints and' + 
                  ' second features')

        transformed_data = np.zeros(shape=data.shape)

        for i in range(data.shape[1]):

            transformed_data[:, i] = data[:, i] * self.mad[i] + self.median[i] 

        return transformed_data


class InvariantScaler(MedianScaler):

    def __init__(self):
        super()

    def fit(self, data):

        super().fit(data)
        
    def fit_transform(self, data):

        self.fit(data)
        return self.transform(data)
    
    def transform(self, data):

        transformed_data = super().transform(data)
        transformed_data = np.arcsinh(transformed_data)

        return transformed_data

    def inverse_transform(self, data):

        transformed_data = np.sinh(data)
        transformed_data = super().inverse_transform(transformed_data)

        return transformed_data


class DataScaler(object):

    """Class to perform data scaling operations

    The scaling technique is defined by the ``normalize`` parameter which takes one of the 
    following values: 

    - ``'Norm'`` for normalizing the data to the interval [0, 1].

    - ``'Norm1'`` for normalizing the data to the interval [-1, 1]. 

    - ``'Std'`` for standarizing the data to follow a normal distribution. 

    - ``'Median'`` for normalizing the data based on the median as defined in as defined in `here <https://doi.org/10.1109/TPWRS.2017.2734563>`_.

    - ``'Invariant'`` for scaling the data based on the asinh transformation (a variance stabilizing transformations) as defined in `here <https://doi.org/10.1109/TPWRS.2017.2734563>`_. 
    
    This class follows the same syntax of the scalers defined in the 
    `sklearn.preprocessing <https://scikit-learn.org/stable/modules/preprocessing.html>`_ module of the 
    scikit-learn library

    Parameters
    ----------
    normalize : str
        Type of scaling to be performed. Possible values are ``'Norm'``, ``'Norm1'``, ``'Std'``, 
        ``'Median'``, or ``'Invariant'``

    Example
    --------
    >>> from epftoolbox.data import read_data
    >>> from epftoolbox.data import DataScaler
    >>> df_train, df_test = read_data(path='.', dataset='PJM', begin_test_date='01-01-2016', end_test_date='01-02-2016')
    Test datasets: 2016-01-01 00:00:00 - 2016-02-01 23:00:00
    >>> df_train.tail()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2015-12-31 19:00:00  29.513832     100700.0      13015.0
    2015-12-31 20:00:00  28.440134      99832.0      12858.0
    2015-12-31 21:00:00  26.701700      97033.0      12626.0
    2015-12-31 22:00:00  23.262253      92022.0      12176.0
    2015-12-31 23:00:00  22.262431      86295.0      11434.0
    >>> df_test.head()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2016-01-01 00:00:00  20.341321      76840.0      10406.0
    2016-01-01 01:00:00  19.462741      74819.0      10075.0
    2016-01-01 02:00:00  17.172706      73182.0       9795.0
    2016-01-01 03:00:00  16.963876      72300.0       9632.0
    2016-01-01 04:00:00  17.403722      72535.0       9566.0
    >>> Xtrain = df_train.values
    >>> Xtest = df_train.values
    >>> scaler = DataScaler('Norm')
    >>> Xtrain_scaled = scaler.fit_transform(Xtrain)
    >>> Xtest_scaled = scaler.transform(Xtest)
    >>> Xtrain_inverse = scaler.inverse_transform(Xtrain_scaled)
    >>> Xtest_inverse = scaler.inverse_transform(Xtest_scaled)
    >>> Xtrain[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])
    >>> Xtrain_scaled[:3,:]
    array([[0.03833877, 0.2736787 , 0.28415155],
           [0.03608228, 0.24425597, 0.24633138],
           [0.03438982, 0.23016409, 0.2261206 ]])
    >>> Xtrain_inverse[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])
    >>> Xtest[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])
    >>> Xtest_scaled[:3,:]
    array([[0.03833877, 0.2736787 , 0.28415155],
           [0.03608228, 0.24425597, 0.24633138],
           [0.03438982, 0.23016409, 0.2261206 ]])
    >>> Xtest_inverse[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])
    """
    
    def __init__(self, normalize):

        if normalize == 'Norm':
            self.scaler = MinMaxScaler(feature_range=(0, 1))
        elif normalize == 'Norm1':
            self.scaler = MinMaxScaler(feature_range=(-1, 1))
        elif normalize == 'Std':
            self.scaler = StandardScaler()
        elif normalize == 'Median':
            self.scaler = MedianScaler()
        elif normalize == 'Invariant':
            self.scaler = InvariantScaler()        

    def fit_transform(self, dataset):
        """Method that estimates an scaler object using the data in ``dataset`` and scales the data in  ``dataset``
        
        Parameters
        ----------
        dataset : numpy.array
            Dataset used to estimate the scaler
        
        Returns
        -------
        numpy.array
            Scaled data
        """

        return self.scaler.fit_transform(dataset)

    def transform(self, dataset):
        """Method that scales the data in ``dataset``
        
        It must be called after calling the :class:`fit_transform` method for estimating the scaler
        Parameters
        ----------
        dataset : numpy.array
            Dataset to be scaled
        
        Returns
        -------
        numpy.array
            Scaled data
        """

        return self.scaler.transform(dataset)

    def inverse_transform(self, dataset):
        """Method that inverse-scale the data in ``dataset``
        
        It must be called after calling the :class:`fit_transform` method for estimating the scaler

        Parameters
        ----------
        dataset : numpy.array
            Dataset to be scaled
        
        Returns
        -------
        numpy.array
            Inverse-scaled data
        """

        return self.scaler.inverse_transform(dataset)

def scaling(datasets, normalize):
    """Function that scales data and returns the scaled data and the :class:`DataScaler` used for scaling.

    It rescales all the datasets contained in the list ``datasets`` using the first dataset as reference. 
    For example, if ``datasets=[X_1, X_2, X_3]``, the function estimates a :class:`DataScaler` object using the array ``X_1``, 
    and transform ``X_1``, ``X_2``, and ``X_3`` using the :class:`DataScaler` object.

    Each dataset must be a numpy.array and it should have the same column-dimensions. For example, if
    ``datasets=[X_1, X_2, X_3]``, ``X_1`` must be a numpy.array of size ``[n_1, m]``,
    ``X_2`` of size ``[n_2, m]``, and ``X_3`` of size ``[n_3, m]``, where ``n_1``, ``n_2``, ``n_3`` can be
    different.

    The scaling technique is defined by the ``normalize`` parameter which takes one of the 
    following values: 

    - ``'Norm'`` for normalizing the data to the interval [0, 1].

    - ``'Norm1'`` for normalizing the data to the interval [-1, 1]. 

    - ``'Std'`` for standarizing the data to follow a normal distribution. 

    - ``'Median'`` for normalizing the data based on the median as defined in as defined in `here <https://doi.org/10.1109/TPWRS.2017.2734563>`_.

    - ``'Invariant'`` for scaling the data based on the asinh transformation (a variance stabilizing transformations) as defined in `here <https://doi.org/10.1109/TPWRS.2017.2734563>`_. 


    The function returns the scaled data together with a :class:`DataScaler` object representing the scaling. 
    This object can be used to scale other dataset using the same rules or to inverse-transform the data.
    
    Parameters
    ----------
    datasets : list
        List of numpy.array objects to be scaled.
    normalize : str
        Type of scaling to be performed. Possible values are ``'Norm'``, ``'Norm1'``, ``'Std'``, 
        ``'Median'``, or ``'Invariant'``
    
    Returns
    -------
    List, :class:`DataScaler`
        List of scaled datasets and the :class:`DataScaler` object used for scaling. Each dataset in the 
        list is a numpy.array.
    
    Example
    --------
    >>> from epftoolbox.data import read_data
    >>> from epftoolbox.data import scaling
    >>> df_train, df_test = read_data(path='.', dataset='PJM', begin_test_date='01-01-2016', end_test_date='01-02-2016')
    Test datasets: 2016-01-01 00:00:00 - 2016-02-01 23:00:00
    >>> df_train.tail()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2015-12-31 19:00:00  29.513832     100700.0      13015.0
    2015-12-31 20:00:00  28.440134      99832.0      12858.0
    2015-12-31 21:00:00  26.701700      97033.0      12626.0
    2015-12-31 22:00:00  23.262253      92022.0      12176.0
    2015-12-31 23:00:00  22.262431      86295.0      11434.0
    >>> df_test.head()
                             Price  Exogenous 1  Exogenous 2
    Date                                                    
    2016-01-01 00:00:00  20.341321      76840.0      10406.0
    2016-01-01 01:00:00  19.462741      74819.0      10075.0
    2016-01-01 02:00:00  17.172706      73182.0       9795.0
    2016-01-01 03:00:00  16.963876      72300.0       9632.0
    2016-01-01 04:00:00  17.403722      72535.0       9566.0
    >>> Xtrain = df_train.values
    >>> Xtest = df_train.values
    >>> [Xtrain_scaled, Xtest_scaled], scaler = scaling([Xtrain,Xtest],'Norm')
    >>> Xtrain[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])
    >>> Xtrain_scaled[:3,:]
    array([[0.03833877, 0.2736787 , 0.28415155],
           [0.03608228, 0.24425597, 0.24633138],
           [0.03438982, 0.23016409, 0.2261206 ]])
    >>> Xtest[:3,:]
    array([[2.5464211e+01, 8.5049000e+04, 1.1509000e+04],
           [2.3554578e+01, 8.2128000e+04, 1.0942000e+04],
           [2.2122277e+01, 8.0729000e+04, 1.0639000e+04]])           
    >>> Xtest_scaled[:3,:]
    array([[0.03833877, 0.2736787 , 0.28415155],
           [0.03608228, 0.24425597, 0.24633138],
           [0.03438982, 0.23016409, 0.2261206 ]])
    >>> type(scaler)
    <class 'epftoolbox.data._wrangling.DataScaler'>

    """

    scaler = DataScaler(normalize)

    for i, dataset in enumerate(datasets):
        if i == 0:
            dataset = scaler.fit_transform(dataset)
        else:
            dataset = scaler.transform(dataset)

        datasets[i] = dataset

    return datasets, scaler